#! /usr/bin/php
<?php
  /*
   * quick-and-dirty script to read hcb_rrd output, as produced by 
   * Eneco's toon thermostat.
   * (hcb_rrd is the round robin database engine of toon, logs energy 
   * consumption etc.)
   * 
   * Works for FW version 3.0.32 and earlier. 
   * Later ones are probably fine as well, but haven't been tested. 
   *
   * To make this work:
   *
   * Find a computer with a working php command line interface.
   *
   * Transfer the files /HCBv2/data/hcb_rrd/*.rra and /HCBv2/data/hcb_rrd/*.dat
   * to some directory on that machine.
   *
   * Transfer the file /HCBv2/config/config_hcb_rrd.xml
   * to some (other) directory on that machine.
   *
   * Edit the two (realtive) paths in this script accordingly.
   *
   * run php: php -f ./hcb_rrd.php
   * (this will produce human-readable .csv files from the database files).
   *
   * If you install php-cli on toon, you can run this script there, and then 
   * there's no need to transfer the database files to another machine. 
   *
   * marcelr, june 5th, 2015.
   */

date_default_timezone_set ( 'Europe/Amsterdam' );

/* 
 * This should point to the directory with .rra and .dat files
 * (files as found in /HCBv2/data/hcb_rrd/*.rra and /HCBv2/data/hcb_rrd/*.dat)
 */

$data_root = './rrd/20150605/';

/* 
 * This should point to the directory with config_hcb_rrd.xml
 * (The file as found in /HCBv2/config/config_hcb_rrd.xml)
 */ 

$xml_root = './rrd/';

$rrd = simplexml_load_file( $xml_root.'config_hcb_rrd.xml');

/* print_r($rrd); */

$i = -1;

while ( isset($rrd->rrdLogger[++$i]->rra) ){

  /* compose data filename */
  
  $dat =  $data_root.$rrd->rrdLogger[$i]->uuid.'.dat';
  
  /* read from XML: useful for cross-checking */ 
  
  $deviceUuid  = $rrd->rrdLogger[$i]->deviceUuid;
  $deviceVar   = $rrd->rrdLogger[$i]->deviceVar;
  $deviceSvc   = $rrd->rrdLogger[$i]->deviceSvc;
  $sampleType  = $rrd->rrdLogger[$i]->sampleType;
  $name        = $rrd->rrdLogger[$i]->name;
  
  /*
   * .dat file layout (little endian):
   *
   * magic number: 17 bytes: hcb_rrd_09082011A
   * int: 4 bytes = strlen bytes for next string: 
   * char:         deviceUuid + \0
   * int: 4 bytes = strlen bytes for next string: 
   * char:         deviceVar + \0
   * int: 4 bytes = strlen bytes for next string: 
   * char:         deviceSvc + \0
   * int: 4 bytes = strlen bytes for next string: 
   * char:         sampleType + \0
   *
   * what follows is repeated for each data set:
   *
   * if data type  is integer:
   * int: 4 bytes, unknown, status flag? (0x0 and 0x7fffffff found so far)
   * int: 4 bytes, unknown, size dependent on data type?
   * int: 4 bytes, unknown, size dependent on data type?
   *
   * if data type is double:
   * double: value
   * double: divider
   *
   * then continue with:
   * int: 4 bytes, unix timestamp, last entry (output of mktime();)
   * int: 4 bytes, unix timestamp, last entry (why twice?)
   * int: 4 bytes, minSamplesPerBin
   * int: 4 bytes = strlen bytes for next string: 
   * char:         binLength + \0.
   *               binLength is ASCII text representing the time interval 
   *               between samples, in seconds.
   * int: 4 bytes, file offset for last entry, endpoint of the circular buffer.
   * int: 4 bytes, number of samples in .rra file
   * int: 4 bytes, unknown
   * int: 4 bytes = strlen bytes for next string: 
   * char:         name + \0
   * int: 4 bytes = strlen bytes for next string: 
   * char:         consolidator + \0
   * 
   */
  
  echo 'dat-file = '.$dat."\n";
  
  /* set flags for data type */
  $integer = strcmp( $sampleType, 'integer' );
  if (!$integer )
    $integer = 1;
  else
    $integer = 0;
  
  $double  = strcmp( $sampleType, 'double' );
  if ( !$double )
    $double = 1;
  else
    $double = 0;
  
  /* open .dat file and read */
  
  $fid_d = fopen( $dat, 'r' );
  
  $magic = fread( $fid_d, 17);
  
  $len = unpack( "i", fread( $fid_d, 4 ) );
  $len = $len[1];
  $deviceUuid_d = fread( $fid_d, $len);
  $len = unpack( "i", fread( $fid_d, 4 ) );
  $len = $len[1];
  $deviceVar_d = fread( $fid_d, $len);
  $len = unpack( "i", fread( $fid_d, 4 ) );
  $len = $len[1];
  $deviceSvc_d = fread( $fid_d, $len);
  $len = unpack( "i", fread( $fid_d, 4 ) );
  $len = $len[1];
  $sampleType_d = fread( $fid_d, $len);
  
  /* repeat until end of file: */
  
  $ncol = 0;
  
  while ( !feof( $fid_d ) ) {

    /* this bit of code can be compressed quite a lot */

    if ( $integer == 1){
      $unk0 = unpack( "i", fread( $fid_d, 4 ) );
      $unk0 = $unk0[1];
      $unk1 = unpack( "i", fread( $fid_d, 4 ) );
      $unk1 = $unk1[1];
      $unk2 = unpack( "i", fread( $fid_d, 4 ) );
      $unk2 = $unk2[1];
    }
    if ( $double == 1 ){
      $value = unpack( "d", fread( $fid_d, 8 ) );
      $value = $value[1];
      $divider = unpack( "d", fread( $fid_d, 8 ) );
      $divider = $divider[1];
    }
    
    $timestamp0 = unpack( "i", fread( $fid_d, 4 ) );
    $timestamp0 = $timestamp0[1];
    $timestamp1 = unpack( "i", fread( $fid_d, 4 ) );
    $timestamp1 = $timestamp1[1];
    $minSamplesPerBin = unpack( "i", fread( $fid_d, 4 ) );
    $minSamplesPerBin = $minSamplesPerBin[1];
    
    $len = unpack( "i", fread( $fid_d, 4 ) );
    $len = $len[1];
    $binLength = intval(fread( $fid_d, $len));
    $offset = unpack( "i", fread( $fid_d, 4 ) );
    $offset = $offset[1];
    $nsamples = unpack( "i", fread( $fid_d, 4 ) );
    $nsamples = $nsamples[1];
    $unk3 = unpack( "i", fread( $fid_d, 4 ) );
    $unk3 = $unk3[1];
    $len = unpack( "i", fread( $fid_d, 4 ) );
    $len = $len[1];
    $ext = fread( $fid_d, $len-1);
    $dum = fread( $fid_d, 1);
    $len = unpack( "i", fread( $fid_d, 4 ) );
    $len = $len[1];
    $consolidator = fread( $fid_d, $len);
    
    /* read actual data from file */
    
    $rra = $data_root.$rrd->rrdLogger[$i]->uuid.'-'.$ext.'.rra';
    $csv = $data_root.$name.'-'.$ext.'.csv';
    
    echo 'rra-file = ',$rra."\n";
    echo 'csv-file = '.$csv."\n";
    
    /* open files  */
    
    $fid_r = fopen( $rra, 'r');
    $fid_w = fopen( $csv, 'w');
    
    /* start time for time axis */
    
    $t0 = $timestamp0 - $binLength * ($nsamples -1 );
    
    for ( $cnt = 0; $cnt < $nsamples; $cnt++) {
      
      /* fill time vector */
      
      $t[$cnt] = $t0 + $binLength * $cnt ;
      
      if ( $integer == 1){
	$dum = unpack( "i", fread( $fid_r, 4 ) );
	$data[$cnt] = $dum[1];
	
	/* remove int-style NaN's */
	
	if ( $data[$cnt] ==0x7fffffff )
	  $data[$cnt] = 'NaN';
	else
	  $data[$cnt] = 0.001 * $data[$cnt];
      }
      if ( $double == 1){
	$dum = unpack( "d", fread( $fid_r, 8 ) );
	$data[$cnt] = $dum[1];
      }
    }
    
    /* repartition data array: */
    
    $data1 = array_slice( $data, $offset, $nsamples - $offset ); 
    $data2 = array_slice( $data, 0, $offset );
    $data = array_merge( $data1, $data2 );
    
    /* .. and write to file */
    
    for ( $cnt = 0; $cnt < $nsamples; $cnt++){
      if ( strcmp($data[$cnt], 'NaN' ) )
	fprintf($fid_w, "%d, %1.3f \n", $t[$cnt], $data[$cnt] );
      else
	fprintf($fid_w, "%d, %s \n", $t[$cnt], $data[$cnt] );
    }
    
    fclose( $fid_r );
    fclose ($fid_w);
    
    /* try to read beyond EOF to break out of the loop */
    
    if ($dum = fread($fid_d, 1 )) {
      fseek( $fid_d, ftell($fid_d) - 1 );
      $ncol = $ncol + 1;
    }
  } 
  fclose( $fid_d );
}

?>
